#include "config.h" 
#include <cmath>
// needed to ensure M_PI defined
#include "geometry.h"
// needed for cmatrix_sincos
#include "cmatrix_complex.h"

namespace libcmatrix {

static const double deg_to_rad=M_PI/180.0;

std::ostream& operator << (std::ostream &ostr,const vector3 &a)
{
  return ostr << '(' << a.x << ", " << a.y << ", " << a.z << ')';
}

std::ostream& operator << (std::ostream &ostr,const spherical &a)
{
  return ostr << '(' << a.r << ", " << a.theta/deg_to_rad << ", " << a.phi/deg_to_rad << ')';
}

double dot(const vector3 &A,const vector3 &B)
{
  return A.x*B.x+A.y*B.y+A.z*B.z;
}

double vector_angle(const vector3 &A,const vector3 &B)
{
  return acos(dot(A,B)/(A.length()*B.length()));
}

vector3 cross(const vector3 &A,const vector3 &B)
{
  return vector3(A.y*B.z -A.z*B.y, -A.x*B.z +A.z*B.x, A.x*B.y -A.y*B.x);
}

// Klauss p. 447
void rotation_matrix(rmatrix3 &R,const Euler &ang)
{
  double ca,sa,cb,sb,cg,sg;

  cmatrix_sincos(ang.alpha,sa,ca);
  cmatrix_sincos(ang.beta,sb,cb);
  cmatrix_sincos(ang.gamma,sg,cg);

  R[0][0]=ca*cb*cg-sa*sg;
  R[0][1]=sa*cb*cg+ca*sg;
  R[0][2]=-sb*cg;

  R[1][0]=-ca*cb*sg-sa*cg;
  R[1][1]=-sa*cb*sg+ca*cg;
  R[1][2]=sb*sg;

  R[2][0]=ca*sb;
  R[2][1]=sa*sb;
  R[2][2]=cb;
}

void vector3::rotate(const rmatrix3 & R)
{
  const double ox=x;
  const double oy=y;

  x=R[0][0]*ox+R[0][1]*oy+R[0][2]*z;
  y=R[1][0]*ox+R[1][1]*oy+R[1][2]*z;
  z=R[2][0]*ox+R[2][1]*oy+R[2][2]*z;
}

vector3 rotate(const vector3 &A,const rmatrix3 &R)
{
  return vector3 (R[0][0]*A.x+R[0][1]*A.y+R[0][2]*A.z,
		 R[1][0]*A.x+R[1][1]*A.y+R[1][2]*A.z,
		 R[2][0]*A.x+R[2][1]*A.y+R[2][2]*A.z);
}

void vector3::rotate(const Euler &rot)
{
  rmatrix3 R;
  rotation_matrix(R,rot);
  rotate(R);
}

vector3 rotate(const vector3 &A,const Euler &rot)
{
  rmatrix3 R;
  rotation_matrix(R,rot);
  return rotate(A,R);
}

vector3::vector3(const spherical &a)
{
  double cphi,sphi,ctheta,stheta;
  cmatrix_sincos(a.theta,stheta,ctheta);
  cmatrix_sincos(a.phi,sphi,cphi);
  const double projr=a.r*stheta;
  x=projr*cphi;
  y=projr*sphi;
  z=a.r*ctheta;
}

spherical::spherical(const vector3 &a)
{
  r=std::sqrt(a.x*a.x+a.y*a.y+a.z*a.z);
  if (r) {
    theta=acos(a.z/r);
    phi=atan2(a.y,a.x);
  }
  else
    theta=phi=0.0;
}

}//namespace libcmatrix
