//Using PDB-file this program calculates root-sum-squared dipolar coupling between specific atoms
//Updates 11/07/2005: for better scalability class ZMolStructure has been introduced
//Updates 13/07/2005: now program can analyse CONECT section and descriminate between inter- and intramolecular coupling
#include <fstream>
#include <string.h>
#include <fnmatch.h>
#include "ttyio.h"
#include "simpsonio.h"
#include "geometry.h"
#include "powder.h"
#include "wigner.h"
#include "cmatrix.h"
#include "List.h"
#include "cmatrix_utils.h"
#include <set>
#include <map>
#include "Histogram.h"
#include "matlabio.h"
#include "NMR.h"

using namespace std;
using namespace libcmatrix;

// Equal steps over alpha and beta
class Hemisphere : public PowderMethod {
public:
  Hemisphere(size_t asteps, size_t bsteps);
  void index_to_orientation(Euler&, double&, size_t) const;
  PowderMethod* clone() const { return new Hemisphere(*this); }

private:
  size_t asteps;
  double astep,bstep;

  double getbeta(size_t whichb) const { return whichb*bstep; }
};

  Hemisphere::Hemisphere(size_t _asteps,size_t _bsteps) : PowderMethod(_asteps*_bsteps,2), asteps(_asteps)
{
  if (asteps<2 || _bsteps<2)
    throw InvalidParameter("Hemisphere()");

  astep=(2*M_PI)/asteps;
  bstep=(0.5*M_PI)/(_bsteps-1);

  double total=0;
  for (size_t i=_bsteps;i--;)
    total+=sin(getbeta(i));
  normal=1.0/(asteps*total);
}

void Hemisphere::index_to_orientation(Euler& dest,double &w,size_t which) const
{
  dest.alpha=astep*(which % asteps);
  dest.beta=getbeta(which / asteps);
  w=normal*sin(dest.beta);
}

const size_t MAX_RECORD=100;
const size_t MAX_TYPE_LEN=16;
bool motionaveraging=false;
size_t nzcw=0;

enum output_t { OUT_NORMAL=0, OUT_TABLES, OUT_MAP, OUT_HIST };
output_t outputtype=OUT_NORMAL;

inline bool isorientationmap() { return (outputtype==OUT_MAP) || (outputtype==OUT_HIST); }

size_t nbeta=1;
size_t nalpha=1;
char outname[256]="";

inline char* chew_null(char* start) {
  char* entry=start;
  while (*entry=='\0')
    entry++;
  return entry;
}

inline void push_ifnonnull(List<size_t>& atomlist, char* start)
{
  const size_t atom=atoi(chew_null(start)); //atom.serial number	
  if (atom)
    atomlist.push_back(atom);
}

//! NB sign of output has no meaning (unlike strcmp)
static int strrcmp(const char* a, const char* b)
{
  const int la=strlen(a);
  const int lb=strlen(b);
  return (lb>la) ? 1 : strcmp(a+la-lb,b);
}

//! slow search
size_t searchnuc(const char* name)
{
  size_t nuc=NULL_NUCLEUS;
  for (size_t i=maximumnuc()-1;i!=NULL_NUCLEUS;i--) {
    if (strrcmp(nuctolabel(i),name)==0) {
      if (nuc) {
	std::cerr << "Nucleus label is ambiguous: " << name << '\n';
	exit(1);
      }
      nuc=i;
    }
  }
  return nuc;	
}

typedef std::map<std::string,size_t> map_t;
static map_t nucmap;

//! cached search
size_t castep2nuc(const char* name)
{
  const std::string strname(name);
  const map_t::iterator iter(nucmap.find(strname));
  return (iter==nucmap.end())
    ? (nucmap[strname]=searchnuc(name))
    : iter->second;
}

const char* get_isotope_name(const char* type)
{
  char buf[8]; 

  size_t i=0; //copy atom type (alphabet chars)
  for (;i<sizeof(buf)-1;i++) {
    if (isalpha(type[i]))
      buf[i]=type[i];
    else
      break;
  }
  buf[i]='\0'; //terminate

//   switch (c) {
//   case 'H': return "1H";
//   case 'C': return "13C";
//   case 'F': return "19F";
// 	case 'P': return "31P";
//   }
//   std::cerr << "Failed to identify nucleus: " << c << std::endl;
//   throw Failed("get_isotope_name");
  return nuctolabel(castep2nuc(buf));
}

template<typename T> bool ismember(const BaseList<T>& vec, const T& v)
{
  for (size_t i=vec.size();i--;) {
    if (vec[i]==v)
      return true;
  }
  return false;
}

typedef std::set<size_t> set_t;

//check that subsel fits exactly within sel
bool checkoverlap(const BaseList<size_t>& sel, const BaseList<size_t>& subsel)
{
  for (size_t i=subsel.size();i--;) {
    if (!ismember(sel,subsel(i))) {
      std::cerr << "Methyl group " << subsel << " not entirely within selection " << sel << ":  results will be bogus\n";
      return false;
    }
  }
  return true;
}

ostream& operator<< (ostream& ostr, const set_t& a)
{
  const set_t::const_iterator end(a.end());
  set_t::const_iterator iter(a.begin());
  bool first=true;
  while (iter!=end) {
    if (first)
      first=false;
    else
      ostr << ',';
    ostr << *iter;
    ++iter;
  }
  return ostr;
}

int verbose=0;
bool asymmerge=false;

struct mol_data { //contains information about an individual atom
vector3 coords; //contain coordinates
char type[MAX_TYPE_LEN]; //contains the atom type, e.g. H or C2
int snumber; //serial number (usually starting from 1)
};

std::ostream& operator<< (std::ostream& ostr, const mol_data& a)
{
  return ostr << a.type << " (" << a.snumber << "): " << a.coords;
}

template<class T> void rearrange(List<T>& dest, const List<T>& source, const BaseList<size_t>& sizes)
{
  const size_t totsizes=sum(sizes);
  const size_t nels=source.size();
  if (nels % totsizes)
    throw Failed("rearrange: number of elements doesn't divide");
  dest.create(nels);
  const size_t cells=nels / totsizes;
  size_t sptr=0;
  size_t offset=0;
  for (size_t div=0;div<sizes.size();div++) {
    size_t destbase=offset;
    const size_t n=sizes(div);
    for (size_t cell=0;cell<cells;cell++) {
      for (size_t i=0;i<n;i++)
	dest(destbase+i)=source(sptr++);
      destbase+=totsizes;
    }
    offset+=n;
  }
}
  
class ZMolStructure {
public:
  ZMolStructure() : cellsize_(0), ncells_(0) {}

  void readPDB(const char*, bool resort =false); //read PDB-file with given filename analysing HETATM/ATOM and CONECT sections
  size_t findSerialNumber(size_t n) const { return serial_to_internal_(n); }//return internal index of atom with supplied serial number
  List<size_t> findAtomType(const char*, bool incell0 =false) const; //return internal indexes of atoms, which type is matched with argument

//   size_t cell0_start() const {
//     if (cell0_==-1)
//       throw Failed("cell size undefined");
//     return cell0_*cellsize_; }

//   size_t cell0_end() const { 
//     if (cell0_==-1)
//       throw Failed("cell size undefined");
//     return (cell0_+1)*cellsize_; }
  const vector3& centre() const { return centre_; }

  size_t whichMolecule(size_t m) const {//return internal index of molecule, which contain atom with given internal index
    if (hasconnections())
      throw Failed("whichMolecule impossible without connection data");
    return whichmolecule_(m);
  }

  bool samemolecule(size_t m, size_t n) const { return (whichMolecule(m)==whichMolecule(n)); }

  void print_atom(size_t, std::ostream&) const;

  bool hasconnections() const { return !(connections.empty()); }
  const mol_data& operator()(size_t i) const { return data(i); }
  size_t size() const { return data.size(); }
  void print(std::ostream& =std::cout) const;

  void write_PDB(const char*) const;

private:
  List<mol_data> data;//contain mol_data for all atoms
  List< List<size_t> > connections;//contain serial numbers, grouped by molecules
  List<size_t> whichmolecule_;
  List<size_t> serial_to_internal_;
  vector3 centre_;
  size_t cellsize_,ncells_;
  //cell0_;
  bool sortedbycell_; //!< true if data is ordered by unit cell
  List<size_t> cell0contents_; //!< list of atoms in "cell0"

  void readPDB_normal(ifstream&); //normal file read (sorted by unit cell)
  void readPDB_resort(ifstream&); //resort file read
  bool readline(ifstream& file, char* buffer, size_t nbuf); //!< read line (true if successful)
  bool cleanline(mol_data&, char* buffer); //!< clean input line (false if empty)

  static bool partcmp(const char*, const char*);//compare how first atom type from input match the second
  void addConnections(const BaseList<size_t>&); //put entries of CONECT section to the connections superlist (see above)
  void tidy_atoms();
  void tidy_connections();
  void clear_connections() { connections.clear(); }
  void clear() { data.clear(); clear_connections(); }
  bool sortedbycell() const { return sortedbycell_; }
};

void ZMolStructure::print_atom(size_t i,std::ostream& ostr =std::cout) const
{
  const mol_data& a((*this)(i));
  ostr << a.type << '('<< a.snumber;
  if (cellsize_) {
    if (sortedbycell_) 
      ostr << ':' << ((i % cellsize_)+1) << ',' << ((i / cellsize_)+1);
    else
      ostr << ':' << ((a.snumber / cellsize_)+1) << ',' << ((a.snumber % cellsize_)+1);
  }
  ostr << ')';
}

void ZMolStructure::print(std::ostream& ostr) const
{
  for (size_t i=0;i<size();i++) {
    if (cellsize_ && ((i % cellsize_)==0))      
      ostr << "----- cell " << ((i/cellsize_)+1) << " -----\n";
    ostr << i << ": " << (*this)(i) << '\n';
  }
}

void ZMolStructure::tidy_atoms()
{
  size_t maxs=0;
  for (size_t i=data.size();i--;) {
    if (data(i).snumber>maxs)
      maxs=data(i).snumber;
  }
  List<size_t>& mutable_index=const_cast<ZMolStructure*>(this)->serial_to_internal_;
  mutable_index.create(maxs+1,size_t(-1));
  vector3 total(0,0,0);
  List<vector3> totals(ncells_,vector3(0,0,0));
  for (size_t i=data.size();i--;) {
    mutable_index(data(i).snumber)=i;
    total+=data(i).coords;
    totals(i / cellsize_)+=data(i).coords;
  }
  centre_=total*(1.0/data.size());
  
  if (sortedbycell_) {
    size_t cell0=0U;
    if (ncells_>1) {
      const double scalef=1.0/cellsize_;
      double mindist=1e30;
      for (size_t i=ncells_;i--;) {
	vector3 lcentre(totals(i)*scalef);
	const double dist=(lcentre-centre_).length();
	if (verbose>1)
	  std::cout << "Centre of cell " << (i+1) << " at " << lcentre << " (distance=" << dist << " A)\n";
	if (dist<mindist) {
	  mindist=dist;
	  cell0=i;
	}
      }
      if (verbose)
	std::cout << "Cell " << (cell0+1) << " is closest to origin at " << centre_ << " A (distance of " << mindist << " A)\n";
    }
    cell0contents_=slice(cell0*cellsize_,cellsize_);
    //   size_t cell0_start() const {
    //     if (cell0_==-1)
    //       throw Failed("cell size undefined");
    //     return cell0_*cellsize_; }
    
    //   size_t cell0_end() const { 
    //     if (cell0_==-1)
    //       throw Failed("cell size undefined");
    //     return (cell0_+1)*cellsize_; }
    
  }
  else { // not sorted by cell
    if (verbose)
      std::cout << "Finding atoms closest to origin at " << centre_ << " A\n";
    cell0contents_.create(cellsize_);
    //!< find atoms closest to cell origin
    for (size_t typei=0;typei<cellsize_;typei++) { //loop over atom types
      double mindist=1e30;
      size_t mini=-1;
      for (size_t atomi=ncells_;atomi--;) {
	const size_t ind=typei*ncells_+atomi;
	const mol_data& curatom(data(ind));
	const vector3 diff(curatom.coords-centre_);
	const double len=diff.norm();
	if (len<mindist) {
	  mindist=len;
	  mini=ind;
	}
      }
      if (verbose)
	std::cout << data(mini) << '\n';
      cell0contents_(typei)=mini;
    }
  }
  //if (verbose)
  //  std::cout << "Indices of 'central' atoms: " << cell0contents_ << '\n';
}

void ZMolStructure::tidy_connections()
{
  if (!hasconnections())
    return;
    
  {
    List<size_t>& mutable_index=const_cast<ZMolStructure*>(this)->whichmolecule_;
    for (size_t m=connections.size();m--;) {
      const List<size_t>& curmol(connections(m));
      for (size_t n=curmol.size();n--;)
	mutable_index(serial_to_internal_(curmol(n)))=m;
    }
  }
}

inline bool globcompare(const char* pattern, const char* str)
{
  switch (fnmatch(pattern,str,FNM_NOESCAPE)) {
  case 0:
    return true;
  case FNM_NOMATCH:
    return false;
  }
  perror("globcompare");
  exit(1);
}

bool ZMolStructure::partcmp(const char* ref, const char* second) //compare atom types from user's input and PDB-file (e.g. C and C2 are the same, F2 and F2A are the same etc.)
{
  const int n=strlen(ref)-1;
  
  if (ref[n]!='*')
    return (strcmp(ref,second)==0);

  char p[MAX_TYPE_LEN];
  strcpy(p,ref);
  p[n]='\0'; //get rid of trailing *

  for (size_t i=0; i<n; i++) {
    if (p[i]!=second[i])
      return false;
  }
  if (isdigit(second[n]) && isdigit(second[n-1])) // e.g. F2 and F21 are different
    return false;
  if (!isdigit(second[n]) && !isdigit(second[n-1]) && (strlen(second)!=n)) // e.g. C and Cu are different
    return false;
  return true;
}

List<size_t> ZMolStructure::findAtomType(const char* type, bool incell0) const
{
  List<size_t> result;
//   const size_t start = incell0 ? cell0_start() : 0;
//   const size_t end = incell0 ? cell0_end() : data.size();
//   for (size_t i=start;i<end;i++) {
//     if (partcmp(type, data(i).type))
//       result.push_back(i);
//   }
  if (incell0) {
    for (size_t i=0;i<cell0contents_.size();i++) {
      const size_t ind=cell0contents_(i);
      if (globcompare(type, data(ind).type))
	result.push_back(ind);
    }
  }
  else {
    for (size_t i=0;i<data.size();i++) {
      if (globcompare(type, data(i).type))
	result.push_back(i);
    }
  }
  return result;
}

void ZMolStructure::addConnections(const BaseList<size_t>& newcon)
{
  for (size_t i=newcon.size();i--;) {  //looking for what molecule to add
    const size_t curatom=newcon(i);
    for (size_t m=connections.size();m--;) {
      List<size_t>& curmol(connections(m));
      if (ismember(curmol,curatom)) {
	for (size_t j=newcon.size();j--;) { //add all new connections
	  if (!ismember(curmol,newcon(j)))
	    curmol.push_back(newcon(j));
	}
	return;
      }
    }
  }
  //molecule is not found i.e. all connections are new
  connections.push_back(newcon);
}

void ZMolStructure::readPDB(const char* fname, bool resort) //parser for PDB-files (from standard guide)
//see www.rcsb.org/pdb/docs/format/pdbguide2.2/guide2.2_frame.html
//the parser is not very clever and can fault if PDB-file slightly non-standard
{
  //  char buffer[MAX_RECORD];
  ifstream file(fname);
  if (!file.is_open())
    throw Failed("readPBD: file open failed");

  clear();
  sortedbycell_=!resort;
  if (resort)
    readPDB_resort(file);
  else
    readPDB_normal(file);
  file.close();
  if (hasconnections() && verbose) {
    std::cout << connections.size() << " molecules found\n";
    if (verbose>1)
      std::cout<< connections<< '\n';
  }
}

struct mol_data_type_order {
  bool operator()(const mol_data& a, const mol_data& b) {
    return (strcmp(a.type,b.type)<0);
  }
};

void ZMolStructure::readPDB_resort(ifstream& file)
{
  char buffer[MAX_RECORD];
  mol_data single_entry;
  List<size_t> atomlist;
  while (readline(file,buffer,sizeof(buffer))) {
    if (cleanline(single_entry,buffer))
      data.push_back(single_entry);
  }
  sort_ip(data,mol_data_type_order()); //!< sort by label
  size_t ncells=0; //!< flag number of cells
  size_t lastreset=0;
  std::cout << "Found " << data.size() << " atoms\n";
  if (verbose>1) {
    FILE* fp=fopen("atomdump","w");
    for (size_t i=0;i<data.size();i++)
      fprintf(fp,"%i\t%s\n",data(i).snumber,data(i).type);
    fclose(fp);
  }
  for (size_t i=1;i<data.size();i++) {
    const char* curtype=data(i).type;
    const bool match=(strcmp(data(lastreset).type,curtype)==0);
    if (ncells==0) {
      if (!match) {
	ncells=i;
	if (verbose)
	  std::cout << "Assuming " << ncells <<  " asymmetric units\n";
	if (data.size() % ncells) {
	  std::cerr << "Number of atoms (" << data.size() << ") is not multiple of number of asymmetric units (" << ncells << ")\n";
	  exit(1);
	}
	lastreset=i;
      }
    }
    else {
      if ((i % ncells)==0) { //at boundary?
	if (match) {
	  std::cerr << "Expecting new atom type at index " << (i+1) << " (serial number " << data(i).snumber << ") but found " << curtype << " again\n";
	  exit(1);
	}
	lastreset=i;
	if (verbose>1)
	  std::cerr << "New atom type: " << data(lastreset).type << '\n';
      }
      else {
	if (!match) {
	  std::cerr << "Expecting repeated atom type " << data(lastreset).type << " at index " << (i+1) << " (serial number " << data(i).snumber << ") but found " << curtype << "\n";
	  exit(1);
	}
      }
    }
  }
  if (ncells==0) {
    ncells_=data.size();
    cellsize_=1;
  }
  else {
    ncells_=ncells;
    cellsize_=data.size() / ncells;
  }
  tidy_atoms();
}

bool ZMolStructure::readline(ifstream& file, char* buffer, size_t nbuf)
{
  if (file.eof())
    return false;
  file.getline(buffer,nbuf);
  return (strcmp(buffer,"ENDM")!=0);
}

bool ZMolStructure::cleanline(mol_data& single_entry, char* buffer)
{
  const size_t len=strlen(buffer);
  if (len<6)
    return false;
  for (size_t i=len;i--;) {
    char& curchr(buffer[i]);
    if (curchr==' ') 
      curchr='\0'; // Substitute all spaces to eol symbols
    else
      curchr=toupper(curchr); //enforce all upper case
  }
  if (strncmp(buffer,"HETATM",6) && strcmp(buffer,"ATOM"))
    return false; //!< no atom data
  if (buffer[13]=='Q')
    return false; //ignore pseudo-atoms

  single_entry.snumber=atoi(chew_null(buffer+6));  //enter (index) (PDB serial number);
  strcpy(single_entry.type,chew_null(buffer+12)); //enter (atom name)
  single_entry.coords.x=atof(chew_null(buffer+30));
  single_entry.coords.y=atof(chew_null(buffer+38));
  single_entry.coords.z=atof(chew_null(buffer+46));
  return (single_entry.coords.x<=9000) || (single_entry.coords.y<=9000) || (single_entry.coords.z<=9000); //ignore XPLOR pseudo atoms
}

void ZMolStructure::readPDB_normal(ifstream& file)
{
  mol_data single_entry;
  List<size_t> atomlist;
  List<size_t> molsizes;
  char buffer[MAX_RECORD];
  
  cellsize_=0;
  size_t lastreset=0;
  while (readline(file,buffer,sizeof(buffer))) { //analyse each string
    if (cleanline(single_entry,buffer)) {      
      const size_t curoffset=data.size()-lastreset;
      if (cellsize_) {
	if (((curoffset % cellsize_)==0) && (strcmp(data(lastreset).type,single_entry.type)!=0)) { //mismatch from expected
	  if (verbose)
	    std::cout << "Additional molecule detected\n";
	  molsizes.push_back(cellsize_);
	  cellsize_=0; //reset cell size
	  lastreset+=curoffset;
	}
      }
      else {
	if (!data.empty() && (strcmp(data(lastreset).type,single_entry.type)==0)) {//starting to repeat entries?
	  cellsize_=curoffset;
	  if (verbose)
	    std::cout << "Repeated label (" << single_entry.type << ") used to determined unit cell size: " << cellsize_ << " atoms\n";
	}
      }
      
      if (verbose>1) {
	if (cellsize_ && ((curoffset% cellsize_)==0))
	  std::cout << "--- cell " << (1+(curoffset / cellsize_)) << " ---\n";
	std::cout << single_entry.snumber << ": " <<  single_entry.type << " at " << single_entry.coords << '\n';
      }

      data.push_back(single_entry);
      if (!strcmp(buffer,"CONECT")) {
	atomlist.create(0); //reset
	push_ifnonnull(atomlist,buffer+6);
	push_ifnonnull(atomlist,buffer+11);
	push_ifnonnull(atomlist,buffer+16);
	push_ifnonnull(atomlist,buffer+21);
	push_ifnonnull(atomlist,buffer+26);
	addConnections(atomlist);
	if (verbose>2)
	  std::cout<< "Connected atoms: " << atomlist << '\n';
      }      
    }
  }
  if (cellsize_==0) {
    cellsize_=data.size()-lastreset;
    ncells_=1;
  }
  if (!molsizes.empty()) { //multiple molecules per cell
    molsizes.push_back(cellsize_);
    if (verbose)
      std::cout << "Molecule sizes: " << molsizes << '\n';
    //re-arrange
    if (hasconnections()) {
      if (verbose)
	std::cout << "Warning: connection information discarded\n";
      clear_connections(); //too much trouble to re-order
    }
    cellsize_=sum(molsizes);
    if (data.size() % cellsize_) {
      std::cerr << data.size() << " atoms does not divide into total cell size of " << cellsize_ << " atoms\n";
      exit(1);
    }
    ncells_ = data.size() / cellsize_;
    List<mol_data> ndata;
    rearrange(ndata,data,molsizes);    
    data.swap(ndata);
    tidy_atoms();
  }    
  else {
    if (cellsize_==0) {
      std::cerr << "Failed to identify unit cell - file corrupt?\n";
      exit(1);
    }
    if (data.size() % cellsize_)
      std::cerr << "Warning: " << data.size() << " atoms does not divide into cell size of " << cellsize_ << " atoms\n";
    else
      ncells_ = data.size() / cellsize_;

    tidy_atoms();
    tidy_connections();
  }
}

struct methyl {
  size_t inds[3];

  methyl(size_t i1, size_t i2, size_t i3) {
    inds[0]=i1;
    inds[1]=i2;
    inds[2]=i3;
  }
  BaseList<size_t> operator()() const { return BaseList<size_t>(3,const_cast<size_t*>(inds)); }
};

inline List<size_t> spec_to_indices(const ZMolStructure& cryst, const char* spec, bool incell0 =false)
{
  return isdigit(spec[0]) ? List<size_t>(1,cryst.findSerialNumber(atoi(spec))) : cryst.findAtomType(spec,incell0);
}

struct indexed_type_order {
  indexed_type_order(const ZMolStructure& crystv) : cryst(crystv) {}

  bool operator()(const size_t a, const size_t b) {
    return (strcmp(cryst(a).type,cryst(b).type)<0);
  }
  
  const ZMolStructure& cryst;
};

void buildlists(List< List<size_t> >& out, const ZMolStructure& cryst, const char* spec)
{
  List<size_t> matchinginds(spec_to_indices(cryst,spec));
  if (matchinginds.empty()) {
    std::cerr << "Failed to find sites matching: " << spec << '\n';
    exit(1);
  }
  if (isdigit(spec[0])) {
    out.create(1U);    
    out.front()=matchinginds;
    return;
  }
  indexed_type_order indexer(cryst);
  if (verbose>1)
    std::cout << "Indices of matching atoms: " << matchinginds << '\n';
  sort_ip(matchinginds,indexer); //!< sorted by label
  if (verbose>1)
    std::cout << "Sorted by group: " << matchinginds << '\n';
  const char* lasttype;
  out.clear();
  List<size_t> tmp;
  bool donecheck=!asymmerge;
  size_t comparelen=8; //!< maximum label length
  for (size_t curind=0;curind<matchinginds.size();curind++) {
    const size_t atomi=matchinginds(curind);
    if ((lasttype!=NULL) && (strncmp(cryst(atomi).type,lasttype,comparelen)==0))
      tmp.push_back(atomi);
    else {
      if (lasttype) {
	if (verbose)
	  std::cout << "Found group of atoms of type " << lasttype << ": " << tmp << '\n'; 
	out.push_back(tmp); // store last list
      }
      lasttype=cryst(atomi).type;
      if (!donecheck) {	
	comparelen=strlen(lasttype)-1;	
	if (!isalpha(lasttype[comparelen]))
	  std::cerr << "Warning: merging asymmetric units and last character is not alphabetic: " << lasttype << '\n';
	donecheck=true;
      }
      tmp.create(1U,atomi);
    }
  }
  if (verbose)
    std::cout << "Found group of atoms of type " << lasttype << ": " << tmp << '\n'; 
  out.push_back(tmp);    
}

template<class T> void accumulate_intersection(T& dest, const T& source)
{
  T tmp;
  set_intersection(dest.begin(),dest.end(),source.begin(),source.end(),insert_iterator<T>(tmp,tmp.begin()));
  dest.swap(tmp);
}

double average_dss(const ZMolStructure& cryst, Matrix<bool>& included, BaseList<bool> atomincluded, const BaseList<size_t>& whichi, const BaseList<size_t>& whichj, rmatrix3& rotmat)
{
  static const double gamma1H(gamma("1H"));

  if (included(whichi.front(),whichj.front()))
    return 0.0;
  atomincluded(whichi)=true;

  const size_t N=whichi.size()*whichj.size();
  double dss=0.0;
  bool includecheck=true;

  if (isorientationmap()) {
    double meand=0.0;
    for (size_t i=whichi.size();i--;) {
      const size_t seli=whichi(i);
      for (size_t j=whichj.size();j--;) {
	const size_t selj=whichj(j);
	const vector3 diff(cryst(seli).coords-cryst(selj).coords);
	const vector3 rotvec(rotate(diff,rotmat));
	const spherical sphco(rotvec);
	if (sphco.r==0.0)
	  throw Failed("spins are co-incident");
	double d=dipolar_coupling(gamma1H,gamma1H,sphco.r*1e-10);
	d*=legendre(2,sphco.theta);
	meand+=d;
	if (includecheck) {
	  if (included(seli,selj))
	    std::cerr << "Warning coupling " << seli << ',' << selj << " seems to be multiply included!\n";
	  included(seli,selj)=included(selj,seli)=true;
	}
      }
    }
    dss+=meand*meand/N;  
    includecheck=false;
  }
  else {
    PlanarZCW powdmeth(nzcw);
    double scale;
    Euler powder;
    
    while (powdmeth.next(powder,scale)) {
      double meand=0.0;
      rotation_matrix(rotmat,powder);
      for (size_t i=whichi.size();i--;) {
	const size_t seli=whichi(i);
	for (size_t j=whichj.size();j--;) {
	  const size_t selj=whichj(j);
	  const vector3 diff=cryst(seli).coords-cryst(selj).coords;
	  const vector3 rotvec(rotate(diff,rotmat));
	  const spherical sphco(rotvec);
	  if (sphco.r==0.0)
	    throw Failed("spins are co-incident");
	  double d=dipolar_coupling(gamma1H,gamma1H,sphco.r*1e-10);
	  d*=legendre(2,sphco.theta);
	  meand+=d;
	  if (includecheck) {
	    if (included(seli,selj))
	      std::cerr << "Warning coupling " << seli << ',' << selj << " seems to be multiply included!\n";
	    included(seli,selj)=included(selj,seli)=true;
	  }
	}
      }
      dss+=scale*meand*meand/N;  
      includecheck=false;
    }  
    static const double scalef=20.0/9;
    dss*=scalef;
  }
  if (verbose)
    cout << "Averaged drss between " << whichi << " and " << whichj << ": " << sqrt(dss) << " Hz\n";
  return dss;
}

void dumptype(FILE* fp, const char* s)
{
  if (asymmerge) //strip asymmetric cell marker
    fwrite(s,1,strlen(s)-1,fp);
  else 
    fprintf(fp,s);
}

//! output data table 
void annotated_table(const Matrix<double>& data, const char* sup, const ZMolStructure& cryst, const BaseList<size_t>& first, const BaseList< List<size_t> >& seconds, bool annot =true)
{
  char buffer[256];
  snprintf(buffer, sizeof(buffer), "%s_%s",outname,sup);
  if (annot) {
    FILE* fp=fopen(buffer,"w");
    for (size_t c=0;c<seconds.size();c++) {
      fprintf(fp,"\t");
      dumptype(fp,cryst(seconds(c).front()).type);//!< output name of first in each (should be all same)
    }
    fprintf(fp,"\n");
    for (size_t r=0;r<first.size();r++) {
      dumptype(fp,cryst(first(r)).type);
      for (size_t c=0;c<data.cols();c++)
	fprintf(fp,"\t%g",data(r,c));
      fprintf(fp,"\n");
    }
    fclose(fp);
  }
  else
    write_matrix(buffer,data);
}

void analyse_molecule(const ZMolStructure& cryst, const char* first_type, const char* second_type, FILE* outfile =NULL)
{  
  const size_t natoms=cryst.size();
  
  const List<size_t> Hlist(cryst.findAtomType("H*"));
  if (verbose>1)
    std::cout << "H atoms: " << Hlist << '\n';
  const size_t nH=Hlist.size();
  const double cutoff=2.0;
  List<set_t> neighbours(nH);
  List<int> reverse(natoms,-1);
  for (size_t i=nH;i--;) {
    reverse(Hlist(i))=i;
    neighbours(i).insert(Hlist(i));
    const vector3 posi(cryst(Hlist(i)).coords);
    for (size_t j=i+1;j<nH;j++) {
      const vector3 posj(cryst(Hlist(j)).coords);
      const double r=(posi-posj).length();
      if (r<cutoff) {
	neighbours(i).insert(Hlist(j));
	neighbours(j).insert(Hlist(i));
      }
    }
  }
  
  List<bool> assigned(natoms,false);
  List<methyl> methyllist;
  List<int> pairedatom(natoms,-1);
  List<int> whichmethyl(natoms,-1);
  for (size_t i=0;i<nH;i++) {
    const set_t& curlist(neighbours(i));
    if (verbose>1)
      cout << "Neighbours of H" << Hlist(i) << ": " << curlist << '\n';
    if (assigned(Hlist(i)) || (curlist.size()==1))
      continue; //already assigned / no neighbours
    
    set_t intersect(curlist);
    const set_t::const_iterator end(curlist.end());
    set_t::const_iterator iter(curlist.begin());
    while (iter!=end) {
      const size_t j=*iter;
      ++iter;
      if (j!=i) //ignore self
	accumulate_intersection(intersect,neighbours(reverse(j)));
    }
    const size_t curn=intersect.size();
    set_t::const_iterator liter(intersect.begin());
    const size_t i1=*liter;
    switch (curn) {
    case 1:
      break;
    case 2: {
      const size_t i2=*(++liter);
	pairedatom(i1)=i2;
	pairedatom(i2)=i1;
	assigned(i1)=true;
	assigned(i2)=true;
    }
      break;
    case 3: {
      const size_t i2=*(++liter);
      const size_t i3=*(++liter);
      const size_t whichm=methyllist.size();
      methyllist.push_back(methyl(i1,i2,i3));
      whichmethyl(i1)=whichm;
      whichmethyl(i2)=whichm;
      whichmethyl(i3)=whichm;
      assigned(i1)=true;
      assigned(i2)=true;
      assigned(i3)=true;
      break;
    }
    default:
      std::cerr << "Unexpected number of neighbours: " << curn << '\n';
    }
  }
  if (nH && verbose) {
    std::cout << "Atom pairings: " << pairedatom << '\n';
    std::cout << "In methyl: " << whichmethyl << '\n';
  }

  const List<size_t> first(spec_to_indices(cryst,first_type,true));      
  const size_t nfirst=first.size(); //number of "from" atoms
  if (nfirst==0) {
    std::cerr << "No observe atoms selected!\n";
    return;
  }
  if (verbose)
    std::cout << "Observed atoms: " << first << '\n'; 
  
  
  if ((*first_type=='H') ^ (*second_type=='H'))
    std::cerr << "Warning: motional averaging not handled for heteronuclear case\n";
  
  const double gam1=gamma(get_isotope_name(cryst(first.front()).type));

  List< List<size_t> > seconds;
  Matrix<double> drsstable,rmintable,dmaxtable;

  if (outputtype==OUT_TABLES) {
    buildlists(seconds,cryst,second_type);
    const size_t ngroups=seconds.size();
    drsstable.create(nfirst,ngroups);
    rmintable.create(nfirst,ngroups);
    dmaxtable.create(nfirst,ngroups);  
  }
  else {
    seconds.create(1U);
    seconds.front()=spec_to_indices(cryst,second_type,false);
  }

  const double standard_short=1.75;
  double d2=0; //sum of d^2
  List<rmatrix> drssmap(nfirst);

  for (size_t groupi=0;groupi<seconds.size();groupi++) { //!< loop over master list
    const List<size_t>& second(seconds(groupi));
    const size_t nsecond=second.size(); //number of "to" atoms
    if (nsecond==0) {
      std::cerr << "No remote atoms selected in set " << (groupi+1) << "\n";
      return;
    }
    if (verbose)
      std::cout << "Remote atoms in set " << (groupi+1) << ": " << second << '\n'; 
    
    const double gam2=gamma(get_isotope_name(cryst(second.front()).type));
  
    List<double> d2local(nfirst);
    List<double> rminlocal(nfirst);
    List<double> dmaxlocal(nfirst);
    Matrix<bool> included(natoms,natoms);
    List<bool> atomincluded(natoms);
    char mol_info[20];
    
    bool ok=true;
    for (size_t j=nsecond;ok && j--;) {
      const int which=whichmethyl(second(j));
      if (which>=0)
	ok=checkoverlap(second,methyllist(which)());
    }
        
    static PowderMethod* powdm=NULL;
    if (!powdm) {
      switch (outputtype) {
      case OUT_MAP:
	powdm=new Hemisphere(nalpha,nbeta);
	break;
      case OUT_HIST:
	powdm=new SphericalZCW(nzcw,hemisphere);
	break;
      default:
	powdm=new PowderSingle;
      }
    }
    const size_t orientations=powdm->orientations();
    if (orientations>1)
      cout << "Orientations: " << orientations << std::endl;
    
    powdm->reset();
    double junkscale;
    Euler powder(0,0,0);
    for (size_t orient=0;orient<orientations;orient++) {
      powdm->index_to_orientation(powder,junkscale,orient);
      
      d2local=0.0;
      rminlocal=0.0;
      included=false;
      atomincluded=false;
      
      const bool needrot=(powder.alpha!=0.0) || (powder.beta!=0.0);
      rmatrix3 rotmat;
      rotation_matrix(rotmat,powder);
      
      for (size_t i=0; i<nfirst; i++){
	const size_t seli=first(i);
	const mol_data& crysti(cryst(seli));
	const vector3 crysti_coords= needrot ? rotate(crysti.coords,rotmat) : crysti.coords;
	if (verbose)
	  cout << '\n';
	if (atomincluded(seli)) {
	  if (verbose)
	    cout << "Skipping " << crysti.type << " (equivalent site already considered)\n";
	  continue;
	}
	const bool ismethyli=(whichmethyl(seli)>=0);
	if (!ok && ismethyli)
	  ok=checkoverlap(first,methyllist(whichmethyl(seli))());
	
	for (size_t j=0; j<nsecond; j++){
	  const size_t selj=second(j);
	  if (seli==selj)
	    continue;
	  const bool ismethylj=(whichmethyl(selj)>=0);
	  const mol_data& crystj(cryst(second(j)));
	  const vector3 crystj_coords= needrot ? rotate(crystj.coords,rotmat) : crystj.coords;
	  const spherical sphco(crysti_coords-crystj_coords);
	  if (cryst.hasconnections())
	    strcpy(mol_info,cryst.samemolecule(first(i),first(j)) ? "(intramolecular)" : "(intermolecular)");
	  else
	    mol_info[0]='\0';
	  
	  double r=0.0;
	  double scale=1.0;
	  double locdss=0.0;
	  size_t weight=1;
	  
	  const bool interesting= motionaveraging ? (ismethyli || ismethylj) : (ismethyli && (whichmethyl(seli)==whichmethyl(selj)));
	  
	  if (interesting) {
	    if (ismethyli && ismethylj) { //both in methyl
	      if (whichmethyl(seli)==whichmethyl(selj)) { //in *same* methyl
		r=standard_short;
		strcpy(mol_info,"(intramethyl)");
		if (motionaveraging)
		  scale=-0.5;
	      }
	      else {
		locdss=average_dss(cryst,included,atomincluded,methyllist(whichmethyl(seli))(),methyllist(whichmethyl(selj))(),rotmat);
		weight=3;
	      }
	    }
	    else { //one in, one out
	      if (ismethyli)
		locdss=average_dss(cryst,included,atomincluded,methyllist(whichmethyl(seli))(),ExplicitList<1,size_t>(selj),rotmat);
	      else
		locdss=average_dss(cryst,included,atomincluded,ExplicitList<1,size_t>(seli),methyllist(whichmethyl(selj))(),rotmat);
	      weight=3;
	    }
	  }
	  else { //no methyl business
	    if (pairedatom(seli)==selj) {
	      r=standard_short;
	      strcpy(mol_info,"(close H-H pair)");
	    }
	    else
	      r=sphco.r;
	  }
      
	  if (r!=0.0) {// averaged d?
	    if ((rminlocal(i)==0.0) || (r<rminlocal(i)))
	      rminlocal(i)=r;
	    
	    double d=scale*dipolar_coupling(gam1,gam2,r*1e-10); //dipolar coupling
	    if (isorientationmap())
	      d*=legendre(2,sphco.theta); //!< include orientation

	    if (verbose) {
	      std::cout << "Distance from ";
	      cryst.print_atom(seli);
	      std::cout << " to ";
	      cryst.print_atom(selj);
	      std::cout << ": " << r<< " A   d=" << d << " Hz " << mol_info << '\n';
	    }
	    locdss=d*d;
	  }
	  d2+=weight*locdss;
	  d2local(i)+=locdss;
	}
	if (d2==0) {
	  std::cerr << "No couplings found!\n";
	  exit(1);
	}
	if (isorientationmap()) {
	  rmatrix& curmap(drssmap(i));
	  if (!curmap) {
	    if (outputtype==OUT_HIST)
	      curmap.create(1,orientations);
	    else
	      curmap.create(nbeta,nalpha); //!< assume that alpha is most rapidly varying
	  }
	  BaseList<double> drow(curmap.row());
	  drow(orient)=sqrt(d2local(i));
	}
	else {
	  const double drss=sqrt(d2local(i));
	  dmaxlocal(i)=rminlocal(i) ? dipolar_coupling(gam1,gam2,rminlocal(i)*1e-10) : 0.0;
	  if (outfile)
	    fprintf(outfile,"%s\t%i\t%f\t%f\n",crysti.type,crysti.snumber,drss,rminlocal(i));
	  else  {
	    std::cout << crysti.type << '(' << crysti.snumber << "): d_rss=" << drss << " Hz";
	    if (rminlocal(i))
	      cout << "  Shortest (rigid) distance=" << rminlocal(i) << " A (d=" << dmaxlocal(i) << " Hz)\n";
	    else
	      cout << '\n';
	  }
	}
      }
    }
    if (outputtype==OUT_TABLES) {
      rmintable(range(),groupi)=rminlocal;
      dmaxtable(range(),groupi)=dmaxlocal;
      for (size_t j=d2local.size();j--;)
	drsstable(j,groupi)=sqrt(d2local(j));
    }
  } //loop over groups
  
  if (isorientationmap()) {
    char tmpname[256];
    matlab_controller ctrl(outname,5);
    for (size_t i=0; i<nfirst; i++) {
      if (!drssmap(i).empty()) {
	snprintf(tmpname,sizeof(tmpname),"drss%i",first(i)+1);
	cout << "Writing drss data for " << (first(i)+1) << '\n';
	ctrl.write(drssmap(i),tmpname);
      }
    }
  }
  else {
    if (outputtype==OUT_TABLES) {
      annotated_table(rmintable,"minr",cryst,first,seconds);
      annotated_table(dmaxtable,"maxd",cryst,first,seconds);
      annotated_table(drsstable,"drss",cryst,first,seconds);
    }
    if (!outfile && !d2 && (nfirst>1)) {
      double dav=sqrt(d2/nfirst);
      std::cout <<"\nAverage d_rss="<< dav << " Hz" << std::endl;
    }
  }
}
  
int main(int argc,const char *argv[])
{
  int count=1;
  try {
    
    cout.precision(5);
    char datafile[128]; //filename of PDB file
    bool interactive=true;
    if (argc!=1) {
      if (argc<4) {
	std::cerr << "Syntax: " << argv[0] << " <observed nuclei> <remote nuclei> <filename> [<filename>...]\nExample nuclei selections: C - all Cs, C1 - carbon 1, C1* - all types of carbon 1 (e.g. C1B)\n";
	return 1;
      }
      interactive=false;
    }
       
     
    if (interactive) {
      char first_type[MAX_TYPE_LEN],second_type[MAX_TYPE_LEN];

      getstring(argc,argv,count,"PDB file? ", datafile,sizeof(datafile));
      const bool resort=!getlogical(argc,argv,count,"Atoms sorted by unit cell? ",true);

      verbose=getint(argc,argv,count, "Verbose level? ", 0,2,0);  

      ZMolStructure cryst;
 
      cryst.readPDB(datafile,resort); //read PDB file
      if (verbose>1) {
	std::cout << "Read in:\n";
	cryst.print();
      }
   
      cout << "Atom type examples: C* - all carbons; C1 - carbons at position 1; C1* - carbons at position 1 (including C1B, C1C, etc)\n";
      getstring(argc,argv,count,"From atom type or serial number (e.g. C*, C15, 124)? ", first_type,sizeof(first_type),"H*");
      
      getstring(argc,argv,count,"To atom type (e.g. N*, N17A)? ",second_type,sizeof(second_type),"H*");
      
      if ((*first_type=='H') && (*second_type=='H'))
	motionaveraging=getlogical(argc,argv,count,"Including motional averaging of XH3? ",true);

      cout << "N - normal output\nT - output data tables\nM - drss map\nH - drss histogram\n";
      outputtype=(output_t)getoption(argc,argv,count,"Output type? ","NTMH",OUT_NORMAL);
      if (outputtype) {
	switch (outputtype) {
	case OUT_TABLES:
	  asymmerge=getlogical(argc,argv,count,"Merge asymmetric units (last character of atom type)? ",false);
	  break;
	case OUT_HIST:
	  nzcw=getint(argc,argv,count,"ZCW number? ");
	  break;
	default:
	  nalpha=getint(argc,argv,count,"alpha angles? ",30);
	  nbeta=getint(argc,argv,count,"beta angles? ",30);
	}
	getstring(argc,argv,count,"Output base filename? ",outname,sizeof(outname));
      }
      else {
	if (motionaveraging)
	  nzcw=getint(argc,argv,count,"ZCW number? ");
      }
      
      analyse_molecule(cryst,first_type,second_type);
    }
    else { //loop through files
      for (size_t i=3;i<argc;i++) {
	if (strlen(argv[i])+4>sizeof(outname))
	  throw Failed("Filename too long");
	strcpy(outname,argv[i]);
	char* ptr=strrchr(outname,'.');
	if (ptr && (ptr>strrchr(outname,'/')))
	  strcpy(ptr+1,"txt"); //place extension after .
	else
	  strcat(outname,".txt");
	  
	std::cout << "Analysing: " << argv[i] << std::endl;

	ZMolStructure cryst;
 
	cryst.readPDB(argv[i]); //read PDB file

	FILE* outfile=fopen(outname,"w");
	if (!outfile)
	  std::cerr << "Failed to open: " << outfile << '\n';
	else {
	  fprintf(outfile,"#Output from %s %s %s %s\n",argv[0],argv[1],argv[2],argv[i]);
	  fprintf(outfile,"#<serial no> <drss/Hz> <r_min/A>\n");
	  analyse_molecule(cryst,argv[1],argv[2],outfile);
	  fclose(outfile);
	}
      }
    }
    
  } catch (MatrixException& exc) {
    cerr << exc;
    return 1;
  }
  return 0;
}
