/* Parse magres file to create spin system */

#include "libcmatrix/ttyio.h"
//#include "space_T.h"
#include <errno.h>
#include "libcmatrix/geometry.h"
#include "libcmatrix/smartptr.h"
#include "libcmatrix/ScratchList.h"
#include "libcmatrix/NMR.h"
#include <map>
#include <string>
#include <set>
#include <sstream>
#include "libcmatrix/MoleculeStructure.h"

using namespace libcmatrix;
using namespace std;

static const double asym_zero_tol=1e-4; // asymmetry set to zero if less than this
static const double rad_to_deg=180.0/M_PI;

static const size_t MAX_LINE=1024;

// Ripped from io.f90 of CASTEP
static const double speed_light_si=299792458.0;
static const double planck_si=6.62606876e-34; // J s
static const double electron_mass_si = 9.10938188e-31; // kg
static const double hbar_si=planck_si/(2*M_PI);
static const double elementary_charge_si = 1.602176462e-19; // C
static const double mu_0_si = 4.0e-7*M_PI;
static const double epsilon_0_si = 1.0/(mu_0_si*speed_light_si*speed_light_si);
static const double fine_structure_si = (elementary_charge_si*elementary_charge_si)/(4.0*M_PI*epsilon_0_si*hbar_si*speed_light_si);
static const double joule = 1.0/(fine_structure_si*fine_structure_si*electron_mass_si*speed_light_si*speed_light_si);
static const double hertz = planck_si*joule;
static const double metre = electron_mass_si*speed_light_si*fine_structure_si/hbar_si;
static const double barn=metre*metre*1e-28;

double match_tolerance=0.0;

int quadorder=2;
int prec=4;

size_t verbose=1;

// inline void fill(BaseList<double> dest, const vector3& a)
// {
//   dest.front()=a.x;
//   dest(size_t(1))=a.y;
//   dest(size_t(2))=a.z;
// }
inline void fill(double d[3], const vector3& a)
{
  d[0]=a.x;
  d[1]=a.y;
  d[2]=a.z;
}

void transformation_matrix(rmatrix3& dest, const vector3& x, const vector3& y, const vector3& z)
{
  fill(dest[0],x);
  fill(dest[1],y);
  fill(dest[2],z);
}

inline void dump(std::ostream& ostr, const double d[3])
{
  ostr << d[0] << "  " << d[1] << "  " << d[2] << '\n';
}

std::ostream& operator<< (std::ostream& ostr, const rmatrix3& a)
{
  dump(ostr,a[0]);
  dump(ostr,a[1]);
  dump(ostr,a[2]);
  return ostr;
}

inline const BaseList<double> aslist(const rmatrix3& a) { return BaseList<double>(9,const_cast<double*>(a[0])); }
inline BaseList<double> aslist(rmatrix3& a) { return BaseList<double>(9,a[0]); }

inline void expand(rmatrix& d, const rmatrix3& a)
{
  d.create(3,3);
  d.row()=aslist(a);
}

enum { ALL=0, SIMPSON_SINGLE, PNMRSIM_SINGLE, PNMRSIM_SUM, TABLES };
int mode=PNMRSIM_SINGLE;

typedef std::map<size_t,std::string> sel_t;
sel_t selnuclei; //!< allowed nuclei (empty if all)

int lico=0;

char* getline(FILE* fp)
{
  static char linebuf[MAX_LINE];
  if (!fgets(linebuf,sizeof(linebuf),fp))
    return NULL;
  lico++;
  return linebuf;
}

void dumpnucleus(FILE* fp, size_t nuc)
{
  fprintf(fp," %s",nuctolabel(nuc));
  const sel_t::const_iterator sel(selnuclei.find(nuc));
  if (sel!=selnuclei.end()) {
    const char* qual=(sel->second).c_str();
    if (*qual)
      fprintf(fp,":%s",qual);
  }
}

void symmetrise(Matrix<double>& A)
{
  if (!issquare(A))
    throw NotSquare("symmetrise");
  for (size_t r=A.rows();r--;) {
    for (size_t c=r;c--;)
      A(r,c)=A(c,r)=0.5*(A(r,c)+A(c,r));
  }
}

//! NB sign of output has no meaning (unlike strcmp)
static int strrcmp(const char* a, const char* b)
{
  const int la=strlen(a);
  const int lb=strlen(b);
  return (lb>la) ? 1 : strcmp(a+la-lb,b);
}

//! slow search
size_t searchnuc(const char* name)
{
  size_t nuc=NULL_NUCLEUS;
  for (size_t i=maximumnuc()-1;i!=NULL_NUCLEUS;i--) {
    if (strrcmp(nuctolabel(i),name)==0) {
      if (selnuclei.find(i)!=selnuclei.end()) //!< if in allowed set, accept automatically
	return i;
      if (nuc)
	return NULL_NUCLEUS; //!< ambiguous label
      nuc=i;
    }
  }
  return nuc;	
}

typedef std::map<std::string,size_t> map_t;
static map_t nucmap;

//! cached search
size_t castep2nuc(const char* name)
{
  const std::string strname(name);
  const map_t::iterator iter(nucmap.find(strname));
  if (iter!=nucmap.end())
    return iter->second;
  const size_t nuc=searchnuc(name);
  if (nuc)
    nucmap[strname]=nuc;
  return nuc;
}
 
typedef std::pair<size_t,const char*> parsed_atom_t;

parsed_atom_t parse_castep_atom(char* name)
{
  char* colon=strchr(name,':');
  if (colon)
    *colon++='\0';
  return parsed_atom_t(castep2nuc(name),colon);
}

Euler matrix_to_euler(const Matrix<double>& R)
{
  static const double tol=1e-10; //!< zero test
  if (!issquare(R) || (R.rows()!=3))
    throw InvalidParameter("matrix_to_euler");

  const double beta=acos(R(2U,2U));
  if ((fabs(R(1U,2U))<tol) && (fabs(R(0U,2U))<tol))
    return Euler(0.0,beta,0.0);//!< sin beta =0 hence alpha, gamma angles are arbitrary
  
  const double alpha=atan2(R(1U,2U),R(0U,2U));
  const double gamma=atan2(R(2U,1U),-R(2U,0U));
  return Euler(alpha,beta,gamma);
}

// template<class Container> struct IndexCompare : public std::binary_function<size_t,size_t,bool> {
//   const Container& data;
//   IndexCompare(const Container& datav) : data(datav) {}
//   bool operator()(size_t m, size_t n) { return (data(m)<data(n)); }
// };

// template<class Container> List<size_t> indexsort(const Container& data)
// {
//   List<size_t> index(slice(0,data.size()),mxflag::temporary);
//   std::sort(index.begin(),index.end(),IndexCompare<Container>(data));
//   return index;
// }

//typedef std::pair<std::string,size_t> nucident;

struct nucident {
  nucident(size_t nucv, const char* labelv, size_t indexv) 
    : nuc(nucv),
      label(labelv ? labelv : ""),
      index(indexv) {}

  nucident(parsed_atom_t spec, size_t indexv)
    : nuc(spec.first),
      index(indexv) {
    if (spec.second)
      label=spec.second;
    else {
      char tmp[10];
      snprintf(tmp,sizeof(tmp)-1,"%i",indexv);
      label=tmp;
    }
  }

  size_t nuc;
  std::string label;
  size_t index;

  bool operator== (const nucident& a) const
  { return (nuc==a.nuc) && (index==a.index) && (label==a.label); }

  bool operator!= (const nucident& a) const
  { return (nuc!=a.nuc) || (index!=a.index) || (label!=a.label); }
};

std::ostream& operator<< (std::ostream& ostr, const nucident& a)
{
  ostr << nuctolabel(a.nuc);
  if (a.label.size())
    ostr << ':' << a.label;
  return ostr << '(' << a.index << ')';
}

class ScaledComparator {
public:
  ScaledComparator(double scalefacv) { scaledtol_=fabs(scalefacv*match_tolerance); }

  bool operator()(double x, double y) const {
    if (scaledtol_)
      return (fabs(x-y)<scaledtol_);
    return (x==y);
  }
private:
  double scaledtol_;
};

struct Comparator {
  bool operator()(double x, double y) const {
    if (match_tolerance) {
      const double compval=std::max(fabs(x),fabs(y))*match_tolerance;
      return compval ? ( fabs(x-y) < compval) : true;
    }
    return (x==y);
  }
};

struct interaction {

  interaction(const Matrix<double>& Av)
    : A(Av), iso(0.0), aniso(0.0), asym(0.0) {}

  interaction()
    : iso(0.0), aniso(0.0), asym(0.0) {}

  void process(bool zeroiso =false);

  enum { ISO =0, ANISO, ASYM, ALPHA, BETA, GAMMA };
  rmatrix A;
  double iso,aniso,asym;
  Euler angles;

  void rotate(const rmatrix&);

  double getvalue(Int2Type<ISO>) const { return iso; }
  double getvalue(Int2Type<ANISO>) const { return aniso; }
  double getvalue(Int2Type<ASYM>) const { return asym; }
  double getvalue(Int2Type<ALPHA>) const { return angles.alpha*rad_to_deg; }
  double getvalue(Int2Type<BETA>) const { return angles.beta*rad_to_deg; }
  double getvalue(Int2Type<GAMMA>) const { return angles.gamma*rad_to_deg; }

  template<int N> void dump(FILE* fp) const { fprintf(fp,"%.*g",prec,getvalue(Int2Type<N>())); }

  void dumpangles(FILE* fp, char sep =' ') const {
    dump<ALPHA>(fp); fputc(sep,fp);
    dump<BETA>(fp); fputc(sep,fp);
    dump<GAMMA>(fp);
  }

  bool operator==(const interaction& a) const { return equalpas(a) && equalangles(a); }
  bool equalpas(const interaction& a) const;
  bool equalangles(const interaction& a) const;
};

class averager {
public:
  averager();
  void add(interaction&);
  void perform();
private:
  List<interaction*> storep;
  List<bool> negate;
  vector3 values;
};

averager::averager()
  : values(0.0,0.0,0.0) {}

void averager::add(interaction& inter)
{
  storep.push_back(&inter);
  double lxx,lyy,lzz;
  negate.push_back(inter.aniso<0.0);
  anisotropy_to_cartesian(lxx,lyy,lzz,inter.iso,fabs(inter.aniso),inter.asym);

  values+=vector3(lxx,lyy,lzz);
}

void averager::perform()
{
  if (storep.size()<2)
    return;
  values/=storep.size();
  double liso,laniso,lasym;
  cartesian_to_anisotropy(liso,laniso,lasym,values.x,values.y,values.z);
  for (size_t i=storep.size();i--;) {
    interaction& curint(*(storep(i)));
    curint.iso=liso;
    curint.aniso=(negate(i) ? -laniso : laniso);
    curint.asym=lasym;
  }
}

void interaction::rotate(const rmatrix& R)
{
  A.unitary_simtrans(R);
}

bool interaction::equalpas(const interaction& a) const
{
  static ScaledComparator areequalasym(1.0);
  static Comparator areequal;
  const bool miso=areequal(iso,a.iso);
  const bool maniso=areequal(aniso,a.aniso);
  const bool masym=areequalasym(asym,a.asym);
  if (verbose>1)
    std::cout << miso << maniso << masym;
  return miso && maniso && masym;
}
  
bool interaction::equalangles(const interaction& a) const
{
  static ScaledComparator areequal(2.0*M_PI);
  return areequal(angles.alpha, a.angles.alpha) && areequal(angles.beta, a.angles.beta) && areequal(angles.gamma, a.angles.gamma);
}


  struct nucdesc {
    enum { CSA =0, QUAD };
    nucdesc(nucident identv, const vector3& positionv, const char* extlabelv =NULL)
      : ident(identv),
	multiplicity(1U),
	position(positionv)
	   //	extlabel(extlabelv)
    { isquad=(spin(ident.nuc).deg()>2); 
      if (extlabelv)
	label=extlabelv;
      else {
	std::ostringstream ostr(std::ostringstream::out);
	ostr << nuctolabel(identv.nuc);
	const char* intlabel=identv.label.c_str();
	if (*intlabel)
	  ostr << ':' << intlabel;
	//	ostr << '_' << identv.index;
	label=ostr.str();
      }
    }

    nucident ident;
    bool isquad;
    smartptr<interaction,false> csap,quadp;
    size_t multiplicity;
    //    const char* extlabel;
    std::string label;
    vector3 position;
    //    const char* label() const { return extlabel ? extlabel : ident.label.c_str(); }
    
    //bool operator== (const nucdesc& a) const { return (ident==a.ident); }
    //bool operator!= (const nucdesc& a) const { return (ident!=a.ident); }
    bool operator== (const nucident& a) const { return (ident==a); }
    bool operator!= (const nucident& a) const { return (ident!=a); }

    size_t nucleus() const { return ident.nuc; }
    size_t index() const { return ident.index; }
    void dump(FILE*, int, const char* ="") const;
    void tableintro(FILE*) const;
    void dumpquadtable(FILE* fp) const;
    void dumpshifttable(FILE* fp) const;

    template<int N> void dump(FILE* fp, Int2Type<CSA>, Int2Type<N> which) const { csap->dump<N>(fp); }
    void dump(FILE* fp, Int2Type<CSA>, Int2Type<interaction::ANISO>, bool printp =true) const {
      csap->dump<interaction::ANISO>(fp);
      if (printp)
	fputc('p',fp);
    }
    void dump(FILE* fp, Int2Type<CSA>, Int2Type<interaction::ISO>, bool printp =true) const {
      csap->dump<interaction::ISO>(fp);
      if (printp)
	fputc('p',fp);
    }    
    template<int N> void dump(FILE* fp, Int2Type<QUAD>, Int2Type<N> which) const { quadp->dump<N>(fp); }

    void rotate(const rmatrix3&, const rmatrix&);
  };

void nucdesc::rotate(const rmatrix3& R3, const rmatrix& R)
{
  position=libcmatrix::rotate(position,R3);
  if (csap.get()) {
    //   std::cout << "Tensor before\n" << A;
    csap->rotate(R);
    //  std::cout << "Tensor after\n" << A;
    csap->process();
  }
  if (quadp.get()) {
    quadp->rotate(R);
    quadp->process(true);
  }
}

struct CompareInteraction : public std::binary_function<interaction,interaction,bool> 
{
  bool operator()(const interaction& a, const interaction& b) const { return (a==b); }
};

struct ComparePAS : public std::binary_function<interaction,interaction,bool> 
{
  bool operator()(const interaction& a, const interaction& b) const { return a.equalpas(b); }
};

template<class Func> struct CompareNuc {
  CompareNuc(const nucdesc& nucv) : nuc(nucv) {}
  bool operator()(const nucdesc& a) const { return isequal(a.csap,nuc.csap,'C') && isequal(a.quadp,nuc.quadp,'Q'); }
  bool isequal(const smartptr<interaction,false>& a, const smartptr<interaction,false>& b, char inter) const
  {
    if ((!a) ^ (!b)) //!< one set and one not?
      return false;
    if (!a)
      return true;
    const bool res=func(*a,*b);
    if (verbose>1) 
      std::cout << inter << ' ' << (res ? "match " : "not matched ");
    return res;
  }  
  Func func;
  const nucdesc& nuc;
};

template<class Func> void compress(List<nucdesc>& atoms)
{
  List<nucdesc> newatoms;
  for (size_t i=0;i<atoms.size();i++) {
    const List<nucdesc>::iterator iter(std::find_if(newatoms.begin(),newatoms.end(),CompareNuc<Func>(atoms(i))));
    if (iter==newatoms.end()) {
      newatoms.push_back(atoms(i));
      if (verbose>1)
	std::cout << "Retaining atom " << (i+1) << "  " << atoms(i).label << '\n';
    }
    else {
      if (verbose>1)
	std::cout << "Merging atom " << (i+1) << "  " << atoms(i).label << '\n';      
      (iter->multiplicity)++; //!< increase multiplicity count
    }
  }	 
  atoms.swap(newatoms);
}

void dump(FILE* fp, const BaseList<size_t>& a)
{
  const bool assum=(mode==PNMRSIM_SUM);
  if (a.size()>1)
    fputc(assum ? '|' : '{',fp);
  for (size_t i=0;i<a.size();i++) {
    if (i)
      fputc(',',fp);
    fprintf(fp,"%i",a(i));
  }
  if (a.size()>1)
    fputc(assum ? '|' : '}',fp);
}
      
template<int Int, int N> void dumplist(FILE* fp, const BaseList<nucdesc>& list, Int2Type<Int> whichi, Int2Type<N> which)
{
  const bool assum=(mode==PNMRSIM_SUM);
  try {
    if (list.size()>1)
      fputc(assum ? '|' : '{',fp);
    for (size_t i=0;i<list.size();i++) {
      if (i)
	fputc(',',fp);
      list(i).dump(fp,whichi,which);
    }
    if (list.size()>1)
      fputc(assum ? '|' : '}',fp);
  } catch (Failed&) { //!< catch failure to de-reference smartptr (missing interaction)
    std::cerr << "Interactions must be defined uniformly for all nuclei in this output mode\n";
    exit(1);
  }
}

template<int Int> void dumplistangles(FILE* fp, const BaseList<nucdesc>& list, Int2Type<Int> whichi)
{  
  dumplist(fp,list,whichi,Int2Type<interaction::ALPHA>()); fputc(' ',fp);
  dumplist(fp,list,whichi,Int2Type<interaction::BETA>()); fputc(' ',fp);
  dumplist(fp,list,whichi,Int2Type<interaction::GAMMA>());
}

void nucdesc::dump(FILE* fp, int ind, const char* intro) const
{
  fprintf(fp,"#%s at (%g,%g,%g) A\n",label.c_str(),position.x,position.y,position.z);
  if (csap.get()) {
    fprintf(fp,"%s\tshift %i ",intro,ind);
    dump(fp,Int2Type<CSA>(),Int2Type<interaction::ISO>()); fputc(' ',fp);
    dump(fp,Int2Type<CSA>(),Int2Type<interaction::ANISO>()); fputc(' ',fp);
    dump(fp,Int2Type<CSA>(),Int2Type<interaction::ASYM>()); fputc(' ',fp);
    csap->dumpangles(fp);
    fputc('\n',fp);
  }
  if (quadp.get()) {
    fprintf(fp,"%s\tquadrupole %i %i ",intro,ind,quadorder);
    dump(fp,Int2Type<QUAD>(),Int2Type<interaction::ANISO>()); fputc(' ',fp);
    dump(fp,Int2Type<QUAD>(),Int2Type<interaction::ASYM>()); fputc(' ',fp);
    quadp->dumpangles(fp);
    fputc('\n',fp);
  }
}

void nucdesc::tableintro(FILE* fp) const
{
//   const char* llabel=extlabel;
//   if (!llabel)
//     llabel=nuctolabel(nucleus());
  fprintf(fp,"%s\t%i\t%i\t",label.c_str(),index(),multiplicity);
}

void nucdesc::dumpquadtable(FILE* fpq) const
{
  if (fpq && quadp.get()) {
    tableintro(fpq);
    dump(fpq,Int2Type<QUAD>(),Int2Type<interaction::ANISO>()); fputc('\t',fpq);
    dump(fpq,Int2Type<QUAD>(),Int2Type<interaction::ASYM>()); fputc('\t',fpq);
    quadp->dumpangles(fpq,'\t');
    fputc('\n',fpq);
  }
} 

void nucdesc::dumpshifttable(FILE* fps) const
{
  if (fps && csap.get()) {
    tableintro(fps);
    dump(fps,Int2Type<CSA>(),Int2Type<interaction::ISO>(),false); fputc('\t',fps);
    dump(fps,Int2Type<CSA>(),Int2Type<interaction::ANISO>(),false); fputc('\t',fps);
    dump(fps,Int2Type<CSA>(),Int2Type<interaction::ASYM>()); fputc('\t',fps);
    csap->dumpangles(fps,'\t');
    fputc('\n',fps);
  }
} 
    
void interaction::process(bool zeroiso)
{
  symmetrise(A);
  Matrix<double> V;
  ScratchList<double,3> eigs(3);
  hermitian_eigensystem(V,eigs,A);
  if (verbose>1)
    std::cout << "Eigenvalues: " << eigs << '\n';
  //  std::cout << "Eigenvectors\n" << V << '\n';  
  ScratchList<double,3> tmpeigs(3);
  const double isotropic=(eigs(0U)+eigs(1U)+eigs(2U))/3.0;
  for (size_t i=3;i--;)
    tmpeigs(i)=fabs(eigs(i)-isotropic);
  //std::cout << "Relative eigenvalues: " << tmpeigs << '\n';
  ScratchList<size_t,3> order(0U,0U,0U);
  indexed_sort(order,tmpeigs);
  std::swap(order(0U),order(1U)); //!< order is y,x,z !
  if (verbose>1)
    std::cout << "Order: " << order << '\n';
  const Matrix<double> Vsorted(V(order,range()));
  const ScratchList<double,3> sortedeigs(eigs(order));
  //  std::cout << "Sorted eigenvalues: " << sortedeigs << '\n';
  //  std::cout << "Sorted eigenvectors\n" << Vsorted << '\n';
  //  interaction out;
  cartesian_to_anisotropy(iso,aniso,asym,sortedeigs(0U),sortedeigs(1U),sortedeigs(2U));
  if (asym<asym_zero_tol)
    asym=0.0;
  if (zeroiso)
    iso=0.0;
  angles=matrix_to_euler(Vsorted);
  if (verbose>1) {
    if (!zeroiso)
      std::cout << "Isotropic: " << iso << "  ";
    std::cout << "Anisotropy: " << aniso << "  Asymmetry: " << asym << '\n';
    std::cout << "Euler angles: " << angles << '\n';
  }
  //  return out;
}

void parse_index_list(List<size_t>& dest, const List<nucdesc>& atoms, char* cptr)
{
  for (;;) {
    char* tokptr=strtok(cptr,",");
    if (tokptr==NULL)
      return;
    cptr=NULL;

    char* colonptr=strchr(tokptr,':');
    size_t nuc=NULL_NUCLEUS;
    unsigned int ind=0;
    
    if (colonptr) {
      *colonptr++='\0';
      if (sscanf(colonptr,"%u",&ind)==1) {
	nuc=isdigit(*tokptr) ? labeltonuc(tokptr) : castep2nuc(tokptr);
	if (nuc==NULL_NUCLEUS) {
	  std::cerr << "Nucleus label is ambiguous: " << tokptr << ".  Use isotope specification e.g. 95Mo\n";
	  exit(1);
	}
      }
    }
    if (nuc==NULL_NUCLEUS) {
      std::cerr << "Failed to parse " << tokptr << " as <element><index>\n";
      exit(1);
    }
    const nucident nucsel(parsed_atom_t(nuc,NULL),ind);      
    
    const nucdesc* matchp=std::find(atoms.begin(),atoms.end(),nucsel);      
    if (matchp==atoms.end()) {
      std::cerr << "Failed to find atom " << tokptr << ':' << colonptr << '\n';
      exit(1);
    }
    dest.push_back(matchp-atoms.begin());
  }
}

int main(int argc, const char* argv[])
{
  const size_t MAX_FILE=1024;
  char fname[MAX_FILE];

  int count=1;
  getstring(argc,argv,count,"Input filename? ",fname,sizeof(fname));

  char outname[MAX_FILE];
  getstring(argc,argv,count,"Output filename? ",outname,sizeof(outname));

  if (strcmp(outname,fname)==0) {
    std::cerr << "Output and input files must be different\n";
    return 1;
  }
    
  char nucspec[MAX_LINE];
  getstring(argc,argv,count,"Comma separated list of nuclei to select (ENTER for all, use 95Mo etc. to specify isotope, O:c for central transition only)? ",nucspec,sizeof(nucspec));
  if (nucspec[0]) {
    char* cptr=nucspec;
    char* tokptr;
    while ((tokptr=strtok(cptr,","))) {
      char* colonptr=strchr(tokptr,':');
      if (colonptr) {
	*colonptr++='\0';
	if (colonptr=='\0') {
	  std::cerr << ": qualifier is empty\n";
	  exit(1);
	}
      }
      const size_t nuc=isdigit(*tokptr) ? labeltonuc(tokptr) : castep2nuc(tokptr);
      if (nuc==NULL_NUCLEUS) {
	std::cerr << "Nucleus label is ambiguous: " << tokptr << ".  Use isotope specification e.g. 95Mo\n";
	exit(1);
      }
      selnuclei[nuc]=colonptr ? colonptr : "";
      cptr=NULL;
    }
  }

  enum { O_ALL =0, O_UNIQUE, O_PAS };
  std::cout << "A - All matching spins\nU - Unique parameters\nP - Unique principal components\n";
  const int outputsel=getoption(argc,argv,count,"Combine entries? ","AUP",O_PAS);
  if (outputsel!=O_ALL)
    match_tolerance=getfloat(argc,argv,count,"Matching tolerance (e.g. 0, 1e-6)? ",0.0); 
  
  std::cout << "A - Spin system of all spins\nS - Single spin spectrum (SIMPSON oriented)\nI - Single spin (pNMRsim individual spectra)\nT - Single spin (pNMRsim total spectrum only)\nX - text tables (e.g. for Excel import)\n";
  mode=getoption(argc,argv,count,"Output type? ","ASITX",mode);
  const bool allspins=(mode==ALL) | (mode==TABLES);  
  typedef std::map< char, List<string> > labelmap_t;
  bool havelabels=false;
  labelmap_t labelmap;
  size_t nlabels=0;
  char pdbname[256];
  getstring(argc,argv,count,"PDB file with labels (if any)? ",pdbname,sizeof(pdbname));
  if (pdbname[0]) {
    MoleculeStructure molstruct(2);
    molstruct.read_PDB(pdbname);
    nlabels=molstruct.size();
    if (nlabels==0) {
      std::cerr << "Failed to read PDB file: " << pdbname << '\n';
      exit(1);
    }
    havelabels=true;
    for (size_t i=0;i<nlabels;i++) {	
      const char* curlabel=molstruct(i).type;
      char curatom=*curlabel;
      if (isdigit(curatom))
	curatom=curlabel[1];
      if ((curatom<'A') && (curatom>'Z'))
	std::cerr << "Warning: Bad label in PDB for atom " << i << ": " << curlabel << '\n';
      else {
	List<string>& curlabels(labelmap[curatom]);
	if (verbose>1) 
	  std::cout << (i+1) << " " << curatom << " " << molstruct(i).type << '\n';
	curlabels.push_back(string(curlabel));
      }
    }
  }
       
  //  assingle=getlogical(argc,argv,count,"Single spin per spectrum? ",true);
  
  FILE* fp=fopen(fname,"r");
  if (!fp) {
    std::cerr << "magres2pNMRsim: failed to open " << fname << '\n';
    return 1;
  }

  typedef std::map<size_t,double> Cqmap_t;
  Cqmap_t Cqmap;
  
  int lico=0;
  enum State { BeforeAtom =1, BeforeData };
  State cstate=BeforeAtom;
	
  typedef List<nucdesc> store_t;
  store_t atoms;
  std::set<size_t> usednuc;
  
  bool havequads=false;
  bool haveshifts=false;

  try {

  nucdesc* curatomp=NULL;
  for (;;) {
    char* cptr;
    if ((cptr=getline(fp))==NULL)
      break;

    char nucleus[MAX_LINE];
    unsigned int ind;
    lico++;
    
    switch (cstate) {
    case BeforeAtom:
      if (strstr(cptr," Barn")) { //!< found quadrupole spec?
	int mass=0;
	double Cq;
	if (sscanf(cptr," %s Isotope %i Q = %lg Barn",nucleus,&mass,&Cq)!=3) {
	  char junk[256];
	  if (sscanf(cptr," %s %s %s Q = %lg Barn",nucleus,junk,junk,&Cq)!=4) {
	    std::cerr << "Failed to parse quadrupole specification: " << cptr;
	    continue;
	  }
	}
	const parsed_atom_t nucspec(parse_castep_atom(nucleus));
	const size_t nuc=nucspec.first;
	if (nuc) {
	  const int mass2=atoi(nuctolabel(nuc)); //!< get mass;
	  if ((mass!=0) && (mass!=mass2))
	    std::cerr << "Mismatch between CASTEP mass (" << mass2 << ") and assumed mass (" << mass << ") g mol^-1.  Specification ignored\n";
	  else {
	    Cq*=barn/hertz; //!< convert from atomic to SI
	    Cqmap[nuc]=Cq;
	    std::cout << "Setting Cq for " << nuctolabel(nuc) << ": " << (Cq*1e-6) << " MHz\n";
	  }
	}
	continue;
      }
      switch (sscanf(cptr,"Atom: %s %ui",nucleus,&ind)) {
      case 2: {
	const parsed_atom_t nucspec(parse_castep_atom(nucleus));
	const size_t nuc=nucspec.first;
	const bool foundmatch=(selnuclei.find(nuc)!=selnuclei.end());
	const bool allow=(nuc!=NULL_NUCLEUS) && (selnuclei.empty() || foundmatch);
	if (allow || (verbose>1)) {
	  std::cout << (allow ? "Found " : "Ignoring ") << nucleus;
	  if (nucspec.second) 
	    std::cout << ':' << nucspec.second;
	  std::cout << ' ' << ind << '\n';
	}	
	if (allow) {
	  const nucident cident(nucspec,ind);
	  const store_t::iterator iter(std::find(atoms.begin(),atoms.end(),cident));
	  if (iter==atoms.end()) {
	    if (verbose>1)
	      std::cout << "(New atom)\n";
	    usednuc.insert(nuc);
	    const char* PDBlabel=NULL;
	    if (nucspec.second)
	      havelabels=true;
	    if (nlabels) {
	      static bool donelabelwarn=false;
	      if ((nucspec.second) && !donelabelwarn) {
		std::cerr << "Labels supplied both in CASTEP file and PDB - overwriting CASTEP label\n";
		donelabelwarn=true;
	      }
	      const labelmap_t::const_iterator iter(labelmap.find(nucleus[0]));
	      if (iter!=labelmap.end()) {
		const List<string>& curlabels(iter->second);
		if (ind<=curlabels.size())
		  PDBlabel=curlabels(ind-1).c_str();
	      }
	      if (!PDBlabel)
		std::cerr << "Failed to find label from PDB for atom " << (ind+1) << ": " << nucleus << ' ' << ind << '\n';		
	    }
	    vector3 pos;
	    char* linebuf;
	    if ((linebuf=getline(fp)) && (linebuf=getline(fp))) {
	      char junkc;
	      int junki;
	      if (sscanf(linebuf,"%c %i Coordinates %lg %lg %lg",&junkc,&junki,&(pos.x),&(pos.y),&(pos.z))!=5)
		std::cerr << "Error parsing coordinates line: " << linebuf << " (ignored)\n";
	    }
	    else {
	      std::cerr << "File ended before coordinates found\n";
	      return 1;
	    }
	    atoms.push_back(nucdesc(cident,pos,PDBlabel));
	    curatomp=&(atoms.back());
	  }
	  else {
	    if (verbose>1)
	      std::cout << "(Atom already encountered)\n";
	    curatomp=&(*iter);
	  }
	  cstate=BeforeData;
	}
      }
	break;
      case 0:
	break;
      default:
	std::cerr << "Failed to parse Atom description at line " << lico << '\n';
	return 1;
      }
      break;
    case BeforeData:
      assert(curatomp);
      if (strncmp(cptr,"TOTAL ",6))
	break;
      cptr+=6;
      smartptr<interaction,false>* curintp=NULL;
      bool ignore=false;
      double scale=1.0;
      bool isquad=false;
      if (strncasecmp(cptr,"Shielding",9)==0) {
	curintp=&(curatomp->csap);
	scale=-1.0; // "shielding" -> "shift";
      }
      else {
	if (strncasecmp(cptr,"tensor",6)==0) {
	  isquad=true;
	  curintp=&(curatomp->quadp);
	  if (!(curatomp->isquad)) {
	    if (verbose>1)
	      std::cerr << "Ignoring quadrupole specification for non-quadrupolar nucleus\n";
	    ignore=true;
	  }
	  else {
	    const Cqmap_t::const_iterator iter(Cqmap.find(curatomp->nucleus()));
	    if (iter==Cqmap.end()) {
	      std::cerr << "Cq for " << nuctolabel(curatomp->nucleus()) << " not set\n";
	      return 1;
	    }
	    scale=iter->second;
	  }
	}
      }
      if (!curintp) {
	std::cerr << "Unrecognised interaction: " << cptr << " at line " << lico << '\n';
	return 1;
      }
      if (!ignore) {
	if (isquad)
	  havequads=true;
	else
	  haveshifts=true;
	Matrix<double> A(3,3);
	for (size_t i=0;i<3;) {
	  char* linebuf;
	  if ((linebuf=getline(fp))==NULL) {
	    std::cerr << "End of input file reading tensor\n";
	    return 1;
	  }
	  BaseList<double> currow(A.row(i));
	  switch (sscanf(linebuf,"%lg %lg %lg",&(currow(0)),&(currow(1)),&(currow(2)))) {
	  case 0: case EOF://!< ignore blank lines
	    break; 
	  case 3:
	    i++;
	    break;
	  default:
	    std::cerr << "Failed to read 3 matrix components at line " << lico << '\n';
	    return 1;
	  }
	}
	A*=scale;
	interaction* intp=new interaction(A);
	curintp->reset(intp);
	intp->process(isquad);
	const space_T A_PAS(spatial_tensor(intp->iso,intp->aniso,intp->asym));
	if (verbose>1)
	  std::cout << (isquad ? 'Q' : 'C') << " tensor in PAS:\n" << A_PAS << '\n';
	const space_T A_MF(rotate(A_PAS,intp->angles));
	if (verbose>1)
	  std::cout << "Tensor in MF:\n" << A_MF << '\n';
	const space_T A_MFd(A2(A));
	if (verbose>1)
	  std::cout << "Tensor in MF (direct):\n" << A_MFd << '\n';      
      }
      cstate=BeforeAtom;
      curatomp=NULL;
      break;
    }
  }  
  fclose(fp);

  if (atoms.empty()) {
    std::cerr << "Did not find any nuclei to save!\n";
    return 1;
  }

  if (verbose)
    std::cout << "Found " << atoms.size() << " matching nuclei\n";

  List<size_t> multiplicity;
  if (outputsel!=O_ALL) {
    if (outputsel==O_PAS)
      compress<ComparePAS>(atoms);
    else
      compress<CompareInteraction>(atoms);
    multiplicity.create(atoms.size());
    int uniqueval=-1;
    for (size_t i=0;i<atoms.size();i++) {
      const size_t mult=atoms(i).multiplicity;
      multiplicity(i)=mult;
      if (uniqueval<0)
	uniqueval=mult;
      else {
	if (uniqueval!=mult)
	  uniqueval=0;
      }
    }
    if (uniqueval>0)
      multiplicity.clear(); //!< unique multiplicity
    if (verbose)
      std::cout << "Nuclei left after merging: " << atoms.size() << '\n';
  }

  if(verbose) {
    for (size_t i=0;i<atoms.size();i++) {
      const nucdesc& curatom(atoms(i));
      std::cout << curatom.ident << " at " << curatom.position << '\n';
    }
  }

  bool showdipolar=false;

  char orientstr[MAX_LINE];
  getstring(argc,argv,count,"Bond vector and atom specifying orientation (e.g. H:6,H:5,C:1) [ENTER for none]? ",orientstr,sizeof(orientstr));  
  if (orientstr[0]) {

    List<size_t> orientindices;
    parse_index_list(orientindices,atoms,orientstr);
    if (verbose)
      std::cout << "Orientation indices: " << orientindices << '\n';

    if (orientindices.size()!=3) {
      std::cerr << "Couldn't parse orientation string as 3 comma-separated identifiers (<element><index>)\n";
      exit(1);
    }

    vector3 ivec(atoms(orientindices.front()).position - atoms(orientindices(size_t(1))).position);
    const double ilength=ivec.length();
    ivec/=ilength;
    const vector3 othervec(atoms(orientindices.front()).position - atoms(orientindices(size_t(2))).position);
    vector3 perp=othervec-dot(ivec,othervec)*ivec;
    perp/=perp.length();
    const double perpcheck=dot(perp,ivec);
    const vector3 third=cross(ivec,perp);
    rmatrix3 rotmat;
    //    transformation_matrix(rotmat,perp,third,ivec); //!< rotate so that ivec is along z, 13C is in xz plane
    transformation_matrix(rotmat,third,ivec,perp); //!< rotate so that ivec is along y, 13C is in yz
    if (verbose) {
      std::cout << "Normalised internuclear vector: " << ivec << " (length: " << ilength << " A)\n";
      std::cout << "Perpendicular vector: " << perp << "  (dot product: " << perpcheck << ")\n";
      std::cout << "Rotation matrix\n" << rotmat;
    }
    Matrix<double> fullrotmat;
    expand(fullrotmat,rotmat);

    for (size_t i=atoms.size();i--;)
      atoms(i).rotate(rotmat,fullrotmat);

    const vector3 ovec=0.5*(atoms(orientindices.front()).position + atoms(orientindices(size_t(1))).position);
    for (size_t i=atoms.size();i--;)
      atoms(i).position-=ovec;
    
    vector3 nivec(atoms(orientindices.front()).position - atoms(orientindices(size_t(1))).position);
    nivec/=nivec.length();
    const vector3 nothervec(atoms(orientindices.front()).position - atoms(orientindices(size_t(2))).position);
    vector3 nperp=nothervec-dot(nivec,nothervec)*nivec;
    nperp/=nperp.length();

    if (verbose) {
      std::cout << "Rotated normalised internuclear vector: " << nivec << '\n';
      std::cout << "Rotated perpendicular vector: " << nperp << '\n';
    }
    showdipolar=true;
  }

  char avstr[MAX_LINE];
  getstring(argc,argv,count,"List of spins to average principal values (e.g. H:6,H:5,C:1) [ENTER for none]? ",avstr,sizeof(avstr));  
  if (avstr[0]) {

    List<size_t> avindices;
    parse_index_list(avindices,atoms,avstr);

    if (avindices.size()<2)
      std::cout << "Less than 2 spins specified! Ignoring.\n";
    else {
      averager qav,csav;

      for (size_t i=avindices.size();i--;) {
	const nucdesc& curatom(atoms(avindices(i)));
	if (curatom.csap.get())
	  csav.add(*(curatom.csap));
	if (curatom.quadp.get())
	  qav.add(*(curatom.quadp));
      }
      csav.perform();
      qav.perform();
    }
  }
  
  char buf[256];
  snprintf(buf,sizeof(buf),"%s_tensors.pdb",outname);
  fp=fopen(buf,"wa");
  fputs("HEADER    DUMMYHEADERINFO\n",fp);
  fprintf(fp,"TITLE     Created by magres2pNMRsim from %s\n",fname);
  List<double> anisopar(6);
  bool usequad=true;
  for (size_t i=0;i<atoms.size();i++) {
    const nucdesc& curatom(atoms(i));
    const vector3& coords(curatom.position);    
    const bool isH=(curatom.ident.nuc==H_NUCLEUS);
    const bool isquad = (usequad && !isH);
    const interaction* intp((isquad ? curatom.quadp : curatom.csap).get());
    char totlabel[8];
    const char* source=curatom.label.c_str();
    while (isdigit(*source))
      ++source;
    size_t j=0;
    if (isH) {
      source++;
      totlabel[j++]='F'; // fake F atom
    }
    for (;j<sizeof(totlabel);) {
      if (*source!=':')
	totlabel[j++]=*source;
      if (*source++=='\0')
	break;
    }
    //snprintf(totlabel,sizeof(totlabel-1),"%s%i",label,nuccount);
    fprintf(fp,"HETATM%5i %-4s UNK 0   1    %8.3f%8.3f%8.3f  1.00  0.00\n",i+1,totlabel,coords.x,coords.y,coords.z);
    if (!intp)
      continue;
    const rmatrix& A(intp->A);
    for (size_t j=3;j--;) 
      anisopar(j)=A(j,j);
    anisopar(size_t(3))=0.5*(A(size_t(0),size_t(1))+A(size_t(1),size_t(0)));
    anisopar(size_t(4))=0.5*(A(size_t(0),size_t(2))+A(size_t(2),size_t(0)));
    anisopar(size_t(5))=0.5*(A(size_t(1),size_t(2))+A(size_t(2),size_t(1)));
    double scale=isquad ? 1e-3 : (isH ? 100 : 10.0);
    if (anisopar(size_t(0))<0)
      anisopar*=-1.0;
    if (!isquad) {
      double iso=(anisopar(0)+anisopar(1)+anisopar(2))/3.0;
      const double aniso=fabs(intp->aniso);
      if (iso>1.5*aniso) {
	const double subtract=iso-aniso;
	anisopar(0)-=subtract;
	anisopar(1)-=subtract;
	anisopar(2)-=subtract;
      }
    }      
    //		siso=(s11+s22+s33)/3
    //		s11=s11-siso;
    //		s22=s22-siso;
    //		s33=s33-siso;
    fprintf(fp,"ANISOU%5i %-4s UNK 0   1  ",i+1,totlabel);
    for (size_t i=0;i<6;i++)
      fprintf(fp,"%7i",int(0.5+scale*anisopar(i)));
    fprintf(fp,"    \n");
  }
  fclose(fp);

  if (mode==TABLES) {
    char buf[256];
    
    if (haveshifts) {
      snprintf(buf,sizeof(buf),"%s_shifts.txt",outname);
      FILE* fpshifts=fopen(buf,"w");
      if (fpshifts==NULL) {
	perror("File open");
	exit(errno);
      }
      fprintf(fpshifts,"Label\tIndex\tMultiplicity\tIsotropic shift (unref)/ ppm\tAnisotropy / ppm\tAsymmetry\talpha\tbeta\tgamma\n");
      for (size_t i=0;i<atoms.size();i++)
	atoms(i).dumpshifttable(fpshifts);
      fclose(fpshifts);
    }

    if (havequads) {
      snprintf(buf,sizeof(buf),"%s_quadrupoles.txt",outname);
      FILE* fpquads=fopen(buf,"w");
      if (fpquads==NULL) {
	perror("File open");
	exit(errno);
      }
      fprintf(fpquads,"Label\tIndex\tMultiplicity\tCq / Hz\tAsymmetry\talpha\tbeta\tgamma\n");
      for (size_t i=0;i<atoms.size();i++)
	atoms(i).dumpquadtable(fpquads);      
      fclose(fpquads);
    }
    return 0;
  }

  fp=fopen(outname,"w");
  fprintf(fp,"#Created by magres2pNMRsim from %s\n",fname);
  if (havelabels) {
    fprintf(fp,"#Site labels:");
    for (size_t i=0;i<atoms.size();i++) {
      const char* label=atoms(i).label.c_str();
      //	fprintf(fp,"%c%s", i ? ' ' : '#', label ? label : "<unknown>");      
      fprintf(fp," %s", *label ? label : "<missing>");
    }
    fputc('\n',fp);
  } 

  fprintf(fp,"\tnuclei");
  const size_t uniquenuc=atoms.front().ident.nuc;
  if (allspins) {
    for (size_t i=0;i<atoms.size();i++)
      dumpnucleus(fp,atoms(i).ident.nuc);
  }
  else {
    if (usednuc.size()!=1) {
      std::cerr << "Must be only one type of nucleus active\n";
      return 1;
    }
    dumpnucleus(fp,uniquenuc);
  }
  fputc('\n',fp);

  switch (mode) {
  case PNMRSIM_SINGLE: case PNMRSIM_SUM:
    if (atoms.front().csap.get()) {
      fputs("\tvariable rawshifts ",fp);
      dumplist(fp,atoms,Int2Type<nucdesc::CSA>(),Int2Type<interaction::ISO>()); fputc(' ',fp);
      fputc('\n',fp);
      fputs("\tshift 1 $rawshifts+$(SHIFT",fp);
      const char* llab=nuctolabel(uniquenuc);
      while (*llab)
	fputc(std::toupper(*llab++),fp);
      fputs(") ",fp);
      dumplist(fp,atoms,Int2Type<nucdesc::CSA>(),Int2Type<interaction::ANISO>()); fputc(' ',fp);
      dumplist(fp,atoms,Int2Type<nucdesc::CSA>(),Int2Type<interaction::ASYM>()); fputc(' ',fp);
      dumplistangles(fp,atoms,Int2Type<nucdesc::CSA>());  
      fputc('\n',fp);
    }
    if (atoms.front().quadp.get()) {
      fprintf(fp,"\tquadrupole 1 %i ",quadorder);
      dumplist(fp,atoms,Int2Type<nucdesc::QUAD>(),Int2Type<interaction::ANISO>()); fputc(' ',fp);
      dumplist(fp,atoms,Int2Type<nucdesc::QUAD>(),Int2Type<interaction::ASYM>()); fputc(' ',fp);
      dumplistangles(fp,atoms,Int2Type<nucdesc::QUAD>());      
      fputc('\n',fp);
    }    
    break;

  case ALL:
    for (size_t i=0;i<atoms.size();i++)
      atoms(i).dump(fp,i+1);
    break;
  case SIMPSON_SINGLE:
    for (size_t i=0;i<atoms.size();i++)
      atoms(i).dump(fp,1,i ? "#" : "");
    break;    
  default:
    throw InternalError("Unknown output mode");
  }

  if (showdipolar) {
    for (size_t i=0;i<atoms.size();i++) {
      const double gammai=spin(atoms(i).nucleus()).gamma();
      for (size_t j=i+1;j<atoms.size();j++) {
	const double gammaj=spin(atoms(j).nucleus()).gamma();
	const vector3 ivec(atoms(i).position-atoms(j).position);
	const spherical sphco(ivec);
	if (sphco.r==0.0)
	  throw Failed("dipolar_parameters: spins are co-incident");
	const double d=dipolar_coupling(gammai,gammaj,sphco.r*1e-10);
	fprintf(fp,"\tdipole %i %i %g 0 %g %g\n",i+1,j+1,d,sphco.theta*rad_to_deg,sphco.phi*rad_to_deg);
      }
    }
  }

  if (multiplicity.size()) {
    if (mode==ALL) {
      fputs("#NB spectra should be adjusted by the following multiplicities to correctly reproduce the intensity pattern\n#",fp);
      dump(fp,multiplicity);
      fputc('\n',fp);
    }
    else {
      fputs("\tvariable scalefactors ",fp);      
      dump(fp,multiplicity);
      fputs("\n#NB ensure that scale $scalefactors is present in pulseq to correctly reproduce the intensity pattern\n",fp);
    }
    std::cout << "Warning: non-uniform multiplicity factors.  See output file for more information\n";
  }
  else
    fputs("\tvariable scalefactors 1\n",fp);

  if ((mode==PNMRSIM_SINGLE) || (mode==PNMRSIM_SUM))
    fprintf(fp,"setenv SAVEFLAGS%s\n",(atoms.size()>1) ? " -sum" : "");

  }  catch (MatrixException& exc) {
    std::cerr << exc;
    return 1;
  }
	    
  fclose(fp);
  return 0;
}
